from typing import Optional
from requests import Response

from schemathesis.models import Case
from schemathesis.hooks import HookContext

from . import get_request_header, original_case_successful, set_request_header


def add_invalid_accept_header(
    context: HookContext, case: Case, response: Response
) -> Optional[Case]:
    if original_case_successful(response):
        set_request_header(case, "Accept", "invalid/accept")
        return case
    else:
        return None


def invalid_accept_header(response: Response, case: Case) -> Optional[bool]:
    accept_header = get_request_header(response.request, "Accept")
    if accept_header == "invalid/accept":
        assert (
            response.status_code == 406
        ), "Requests including Accept header with only unsupported formats MUST be rejected with a 406 status code. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#negotiation-headers"
    else:
        # skips the check when it's not relevant
        return True
    return None



from typing import Optional
from requests import Response

from schemathesis.models import Case
from schemathesis.hooks import HookContext

from . import get_request_header, original_case_successful, set_request_header


def add_invalid_request_content_type(
    context: HookContext, case: Case, response: Response
) -> Optional[Case]:
    if original_case_successful(response):
        set_request_header(case, "Content-Type", "invalid/content/type")
        return case
    else:
        return None


def invalid_request_content_type(response: Response, case: Case) -> Optional[bool]:
    if (
        response.request.body
        and get_request_header(response.request, "Content-Type") == "invalid/content/type"
    ):
        assert (
            response.status_code == 415
        ), "415 status code must be used when the client sends a payload with a content-type not supported by the server. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#representation-headers"
    else:
        # skips the check when it's not relevant
        return True
    return None



from requests import PreparedRequest, Response

from schemathesis.models import Case


def get_request_header(request: PreparedRequest, header_name: str) -> str:
    return (
        request.headers[header_name]
        if request and request.headers and header_name in request.headers
        else ""
    )


def original_case_successful(response: Response) -> bool:
    return 200 <= response.status_code < 300


def set_request_header(case: Case, header_name: str, header_val: str) -> None:
    """Adds or overwrites header with header_name."""
    if not case.headers:
        case.headers = {header_name: header_val}
        return

    current_headers_dict = {header.lower(): header for header in case.headers}
    if header_name.lower() in current_headers_dict:
        # overwite the existing header
        case.headers[current_headers_dict[header_name.lower()]] = header_val
    else:
        case.headers[header_name] = header_val



from typing import Any, Callable, Dict, FrozenSet, IO, Iterable, Tuple

import os
import json
import yaml

CONFIG_FILE_NAME: str = "ibm-service-validator-config"
HANDBOOK_CONFIG_NAME: str = "ibm_cloud_api_handbook"
SCHEMATHESIS_CONFIG_NAME: str = "schemathesis_checks"
DEFAULT_CONFIG: Dict[str, Dict[str, str]] = {
    HANDBOOK_CONFIG_NAME: {
        "allow_header_in_405": "on",
        "content_location": "warn",
        "get_with_request_body": "on",
        "invalid_accept_header": "on",
        "invalid_request_content_type": "on",
        "location_201": "on",
        "no_422": "on",
        "no_accept_header": "on",
        "no_content_204": "on",
        "www_authenticate_401": "on",
    },
    SCHEMATHESIS_CONFIG_NAME: {
        "not_a_server_error": "on",
        "status_code_conformance": "on",
        "content_type_conformance": "on",
        "response_schema_conformance": "on",
    },
}


def process_config() -> Tuple[FrozenSet[str], FrozenSet[str]]:
    config = load_config_file_as_dict(os.getcwd())

    if not config:
        config = DEFAULT_CONFIG

    return checks_on(config), warnings(config)


def load_config_file_as_dict(dir: str) -> Dict[str, Any]:
    """Searches up the directory structure for config file and returns config as dict."""

    if not dir:
        return dict()

    path_to_file = os.path.join(dir, CONFIG_FILE_NAME)

    if os.path.exists(path_to_file + ".yaml"):
        return open_get_data_close(path_to_file + ".yaml", yaml.safe_load)
    elif os.path.exists(path_to_file + ".yml"):
        return open_get_data_close(path_to_file + ".yml", yaml.safe_load)
    elif os.path.exists(path_to_file + ".json"):
        return open_get_data_close(path_to_file + ".json", json.load)
    else:
        # search in parent directory
        return load_config_file_as_dict(dir.rsplit("/", 1)[0])


def checks_on(config: Dict[str, Any]) -> FrozenSet[str]:
    """Returns all checks that are not off.

    Note: warnings are included because we still want to run the warning checks.
    """
    # YAML treats unquoted off as implicit boolean. Hence, we check "not x".
    is_off = lambda x: x == "off" or not x
    not_on = {
        *_get_checks(config.get(HANDBOOK_CONFIG_NAME), is_off),
        *_get_checks(config.get(SCHEMATHESIS_CONFIG_NAME), is_off),
    }
    return frozenset(
        {
            *(
                rule
                for rule in DEFAULT_CONFIG[HANDBOOK_CONFIG_NAME]
                if rule not in not_on
            ),
            *(
                rule
                for rule in DEFAULT_CONFIG[SCHEMATHESIS_CONFIG_NAME]
                if rule not in not_on
            ),
        }
    )


def warnings(config: Dict[str, Any]) -> FrozenSet[str]:
    """Returns set of all warnings."""

    is_warning = lambda x: x == "warn"
    return frozenset(
        {
            *_get_checks(config.get(HANDBOOK_CONFIG_NAME), is_warning),
            *_get_checks(config.get(SCHEMATHESIS_CONFIG_NAME), is_warning),
        }
    )


def _get_checks(config: Any, condition: Callable[[Any], bool]) -> Iterable[str]:
    if not config or not isinstance(config, dict):
        return ()

    # YAML treats unquoted off as implicit boolean. Hence, we check "not val".
    return (rule for rule, val in config.items() if condition(val))


def create_default_config(write_json: bool, overwrite: bool) -> None:
    if write_json:
        _create_default_config(CONFIG_FILE_NAME + ".json", overwrite, json.dump)
    else:
        _create_default_config(CONFIG_FILE_NAME + ".yaml", overwrite, yaml.safe_dump)


def _create_default_config(
    file_name: str, overwrite: bool, dump: Callable[[Any, IO], None]
) -> None:
    if os.path.exists(file_name) and not overwrite:
        print(
            "Config file already exists. Use -o or --overwrite to overwrite the config with default values."
        )
        return

    open_write_close(file_name, DEFAULT_CONFIG, dump)


def open_get_data_close(
    path_to_file: str, load: Callable[[IO], Dict[str, Any]]
) -> Dict[str, Any]:
    f = open(path_to_file, "r")
    data = load(f)
    f.close()
    return data


def open_write_close(
    file_name: str, config: Dict[str, Any], dump: Callable[[Any, IO], None]
) -> None:
    """Writes config data to a file in the cwd."""
    config_file = open(file_name, "a+")
    config_file.truncate(0)  # clear contents of file
    dump(config, config_file)
    config_file.close()



from typing import Optional

from requests import Response
from schemathesis.models import Case


def allow_header_in_405(response: Response, case: Case) -> Optional[bool]:
    if response.status_code == 405:
        # response.headers object is case-insensitive
        assert (
            response.headers and "Allow" in response.headers
        ), "For 405 response, must provide the Allow header with the list of accepted request methods. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#negotiation-headers"
    else:
        # skips the test when it's not relevant
        return True
    return None


def content_location(response: Response, case: Case) -> Optional[bool]:
    if response.status_code == 200 and case.method.upper() in {"PUT", "PATCH"}:
        assert (
            response.headers and "Content-Location" in response.headers
        ), "For successful PUT or PATCH response, should provide Content-Location header with the URI of the resource. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#content-location"

        content_location_header = response.headers.get("Content-Location")
        relative_uri = case.formatted_path
        absolute_uri = case._get_base_url() + relative_uri
        assert content_location_header in {
            relative_uri,
            absolute_uri,
        }, f"Content-Location header should match the request URI. Content-Location: {content_location_header}, Absolute Request URI: {absolute_uri}, Relative Request URI: {relative_uri}. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#content-location"
    elif response.status_code == 201:
        assert (
            response.headers and "Content-Location" in response.headers
        ), "For 201 response, should provide Content-Location header with the URI of the created resource. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#content-location"

        content_location_header = response.headers.get("Content-Location")
        location = response.headers.get("Location")
        assert (
            content_location_header == location
        ), f"Content-Location header should match the value of the Location header. Content-Location: {content_location_header}, Location: {location}. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#content-location"
    elif response.status_code == 202 and case.method.upper() in {
        "DELETE",
        "PATCH",
        "POST",
        "PUT",
    }:
        assert (
            response.headers and "Content-Location" in response.headers
        ), "For 202 response from a DELETE, PATCH, POST, or PUT, should provide Content-Location header with the URI where resource may be obtained. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-headers#content-location"
    else:
        # skips the test when it's not relevant
        return True
    return None


def location_201(response: Response, case: Case) -> Optional[bool]:
    if response.status_code == 201:
        assert (
            response.headers and "Location" in response.headers
        ), "For 201 response, must provide Location header with the URI of the created resource. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-status-codes#success-2xx"
    else:
        # skips the test when it's not relevant
        return True
    return None


def no_accept_header(response: Response, case: Case) -> Optional[bool]:
    request_headers = response.request.headers
    if response.content and (not request_headers or "Accept" not in request_headers):
        assert (
            response.headers
            and "Content-Type" in response.headers
            and response.headers["Content-Type"].startswith("application/json")
        ), "Accept header not provided in the request. Response must be JSON, and Content-Type header must start with application/json. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-format#content-type-behavior"
    else:
        # skips the test when it's not relevant
        return True
    return None


def www_authenticate_401(response: Response, case: Case) -> Optional[bool]:
    if response.status_code == 401:
        assert (
            response.headers and "WWW-Authenticate" in response.headers
        ), "401 response must have WWW-Authenticate header. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-status-codes#client-errors-4xx"
    else:
        # skips the test when it's not relevant
        return True
    return None



from ibm_service_validator.handbook_rules.add_case_rules import (
    get_with_request_body,
    invalid_accept_header,
    invalid_request_content_type,
)
from ibm_service_validator.handbook_rules.general_rules import header_rules
from ibm_service_validator.handbook_rules.general_rules import status_code_rules

HANDBOOK_RULES: tuple = (
    get_with_request_body.get_with_request_body,
    invalid_accept_header.invalid_accept_header,
    invalid_request_content_type.invalid_request_content_type,
    header_rules.allow_header_in_405,
    header_rules.content_location,
    header_rules.location_201,
    header_rules.no_accept_header,
    header_rules.www_authenticate_401,
    status_code_rules.no_422,
    status_code_rules.no_content_204,
)

ADD_CASE_HOOKS: tuple = (
    get_with_request_body.add_get_with_request_body,
    invalid_accept_header.add_invalid_accept_header,
    invalid_request_content_type.add_invalid_request_content_type,
)



from typing import Optional

from requests import Response
from schemathesis.models import Case


def no_422(response: Response, case: Case) -> Optional[bool]:
    assert (
        response.status_code != 422
    ), "422 code should not be used. Instead, 400 should be returned in response to invalid request payloads. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-status-codes#client-errors-4xx"
    return None


def no_content_204(response: Response, case: Case) -> Optional[bool]:
    if response.status_code == 204:
        assert (
            not response.content
        ), "204 response must not include a response body. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-status-codes#success-2xx"
    else:
        # skips the test when it's not relevant
        return True
    return None



from typing import Optional
from requests import Response

from schemathesis.models import Case
from schemathesis.hooks import HookContext

from . import original_case_successful


def add_get_with_request_body(
    context: HookContext, case: Case, response: Response
) -> Optional[Case]:
    if original_case_successful(response) and case.method.upper() == "GET":
        # add request body if the original case received 2xx response
        case.body = {"requestBody": "request body passed with GET."}
        return case
    else:
        # otherwise, do not create an additional test case
        return None


def get_with_request_body(response: Response, case: Case) -> Optional[bool]:
    """Validates that request body is ignored for otherwise valid GET request.

    Assumes the constructed request is valid aside from the request body.
    """
    request = response.request
    if request.method == "GET" and request.body:
        assert (
            200 <= response.status_code < 300
        ), "Request body with a GET request must not cause an error. https://cloud.ibm.com/docs/api-handbook?topic=api-handbook-methods#get"
    else:
        # skips the test when it's not relevant
        return True
    return None



from typing import Dict, FrozenSet, List, Optional, Set, Tuple, Union, cast

import click
from schemathesis.cli.context import ExecutionContext
from schemathesis.cli.handlers import EventHandler
from schemathesis.cli.output import default
from schemathesis.models import Status
from schemathesis.runner import events
from schemathesis.runner.serialization import (
    SerializedCase,
    SerializedCheck,
    SerializedTestResult,
)


def handle_after_execution(
    context: ExecutionContext, event: events.AfterExecution, warnings: FrozenSet[str]
) -> None:
    """Display the execution result + current progress at the same line with the method / endpoint names."""
    context.endpoints_processed += 1
    context.results.append(event.result)
    display_execution_result(context, event, warnings)
    default.display_percentage(context, event)


def handle_finished(
    context: ExecutionContext,
    event: events.Finished,
    warnings: FrozenSet[str],
    statistics: bool = False,
) -> None:
    """Show the outcome of the whole testing session."""
    click.echo()
    default.display_hypothesis_output(context.hypothesis_output)
    display_exceptions(context, event)
    display_warnings(context, event, warnings)
    display_errors(context, event, warnings)
    default.display_application_logs(context, event)
    display_totals(context, event, warnings)
    click.echo()
    display_summary(event, warnings, statistics)


def handle_internal_error(context: ExecutionContext, event: events.InternalError) -> None:
    display_internal_error(context, event)
    raise click.Abort


def display_internal_error(
    context: ExecutionContext, event: events.InternalError
) -> None:
    click.secho(event.message, fg="red")
    if event.exception:
        if context.show_errors_tracebacks:
            message = f"Error: {event.exception_with_traceback}\n"
        else:
            message = (
                f"Error: {event.exception}\n"
                f"Add this option to your command line parameters to see full tracebacks: --show-exception-tracebacks"
            )
        if event.exception_type == "jsonschema.exceptions.ValidationError":
            message += (
                "\n\nYou can disable input schema validation with --validate-schema=false "
                "command-line option.\nIn this case, we cannot guarantee proper"
                " behavior during the test run."
            )
        click.secho(
            message, fg="red",
        )


def display_exceptions(context: ExecutionContext, event: events.Finished) -> None:
    """Display all exceptions in the test run."""
    if not event.has_errors:
        return

    default.display_section_name("EXCEPTIONS")
    for result in context.results:
        if result.has_errors:
            display_single_exception(context, result)
    if not context.show_errors_tracebacks:
        click.secho(
            "Add this option to your command line parameters to see full tracebacks: --show-exception-tracebacks",
            fg="magenta",
        )


def display_single_exception(
    context: ExecutionContext, result: SerializedTestResult
) -> None:
    default.display_subsection(result, "magenta")
    for error in result.errors:
        if context.show_errors_tracebacks:
            message = error.exception_with_traceback
        else:
            message = error.exception
        click.secho(message, fg="magenta")
        if error.example is not None:
            display_example(error.example, "magenta", seed=result.seed)


def display_execution_result(
    context: ExecutionContext, event: events.AfterExecution, warnings: FrozenSet[str]
) -> None:
    """Display an appropriate symbol for the given event's execution result.

    Note: The ordering of the conditional blocks affects behavior.
    """
    symbol: str
    color: str
    if event.status == Status.success:
        symbol, color = ".", "green"
    elif event.result.checks and warn_or_success(event.result, warnings):
        symbol, color = "W", "yellow"
    elif event.status == Status.failure:
        symbol, color = "E", "red"
    else:
        # an exception occurred
        symbol, color = "E", "magenta"
    context.current_line_length += len(symbol)
    click.secho(symbol, nl=False, fg=color)


def display_errors(
    context: ExecutionContext, event: events.Finished, warnings: FrozenSet[str]
) -> None:
    """Display all errors in the test run."""
    if not event.has_failures and not event.has_errors:
        return
    errors = [
        result
        for result in context.results
        if result.has_failures and not warn_or_success(result, warnings)
    ]
    if errors:
        default.display_section_name("ERRORS")
        for result in errors:
            display_single_test(get_unique_errors(result.checks, warnings), result, "red")


def display_warnings(
    context: ExecutionContext, event: events.Finished, warnings: FrozenSet[str]
) -> None:
    """Display all warnings in the test run."""
    if not event.has_failures:
        return
    warning_results = [
        result for result in context.results if contains_warning(result, warnings)
    ]
    if warning_results:
        default.display_section_name("WARNINGS")
        for result in warning_results:
            display_single_test(
                get_unique_warnings(result.checks, warnings), result, "yellow"
            )


def display_single_test(
    unique_checks: List[SerializedCheck], result: SerializedTestResult, color: str
) -> None:
    """Display an error or warning for a single method / endpoint."""
    default.display_subsection(result, color)
    display_all_checks(unique_checks, color, result.seed)


def display_all_checks(checks: List[SerializedCheck], color: str, seed: int) -> None:
    for idx, check in enumerate(checks, 1):
        message: Optional[str]
        if check.message:
            message = f"{idx}. {check.message}"
        else:
            message = None
        example = cast(SerializedCase, check.example)
        display_example(example, color, check.name, message, seed)
        # Display every time except the last check
        if idx != len(checks):
            click.echo("\n")


def display_example(
    case: SerializedCase,
    color: str,
    check_name: Optional[str] = None,
    message: Optional[str] = None,
    seed: Optional[int] = None,
) -> None:
    if message is not None:
        click.secho(message, fg=color)
        click.echo()
    output = {
        make_verbose_name(attribute): getattr(case, attribute)
        for attribute in (
            "path_parameters",
            "headers",
            "cookies",
            "query",
            "body",
            "form_data",
        )
    }
    max_length = max(map(len, output))
    template = f"{{:<{max_length}}} : {{}}"
    if check_name is not None:
        click.secho(template.format("Check", check_name), fg=color)
    for key, value in output.items():
        if (key == "Body" and value is not None) or value not in (None, {}):
            click.secho(template.format(key, value), fg=color)
    click.echo()
    click.secho(
        f"Run this Python code to reproduce this failure: \n\n    {case.requests_code}",
        fg=color,
    )
    if seed is not None:
        click.secho(
            f"\nOr add this option to your command line parameters: --hypothesis-seed={seed}",
            fg=color,
        )


def display_totals(
    context: ExecutionContext, event: events.Finished, warnings: FrozenSet[str]
) -> None:
    """Format and print statistic collected by :obj:`models.TestResult`."""
    default.display_section_name("SUMMARY")
    click.echo()
    total = event.total
    if event.is_empty or not total:
        click.secho("No checks were performed.", bold=True)

    if total:
        display_checks_totals(total, warnings)

    if context.cassette_file_name or context.junit_xml_file:
        click.echo()

    if context.cassette_file_name:
        category = click.style("Network log", bold=True)
        click.secho(f"{category}: {context.cassette_file_name}")

    if context.junit_xml_file:
        category = click.style("JUnit XML file", bold=True)
        click.secho(f"{category}: {context.junit_xml_file}")


def display_checks_totals(
    total: Dict[str, Dict[Union[str, Status], int]], warnings: FrozenSet[str]
) -> None:
    padding = 20
    col1_len = max(map(len, total.keys())) + padding
    col2_len = (
        len(str(max(total.values(), key=lambda v: v["total"])["total"])) * 2 + padding
    )
    col3_len = padding
    template = f"    {{:{col1_len}}}{{:{col2_len}}}{{:{col3_len}}}"

    warning_checks = [
        (check_name, results)
        for check_name, results in total.items()
        if check_name in warnings
    ]
    if warning_checks:
        click.secho("Warning checks:", bold=True)
        for check_name, results in warning_checks:
            display_check_result(check_name, results, template, warnings)
        click.echo()

    regular_checks = [
        (check_name, results)
        for check_name, results in total.items()
        if check_name not in warnings
    ]
    if regular_checks:
        click.secho("Performed checks:", bold=True)
        for check_name, results in regular_checks:
            display_check_result(check_name, results, template, warnings)


def display_check_result(
    check_name: str,
    results: Dict[Union[str, Status], int],
    template: str,
    warnings: FrozenSet[str],
) -> None:
    """Show the summary for a single check."""
    if check_name in warnings and Status.failure in results:
        verdict = "WARNING"
        color = "yellow"
    elif Status.failure in results:
        verdict = "ERROR"
        color = "red"
    else:
        verdict = "SUCCESS"
        color = "green"
    success = results.get(Status.success, 0)
    total = results.get("total", 0)
    click.echo(
        template.format(
            check_name,
            f"{success} / {total} passed",
            click.style(verdict, fg=color, bold=True),
        )
    )


def display_summary(
    event: events.Finished, warnings: FrozenSet[str], statistics: bool = False
) -> None:
    counts = get_summary_counts(event, warnings)
    message, color, status_code = get_summary_output(counts, event, warnings)
    if statistics:
        display_statistical_summary(counts, event, warnings, color)
    default.display_section_name(message, fg=color)
    raise click.exceptions.Exit(status_code)


def display_statistical_summary(
    counts: Dict[str, int],
    event: events.Finished,
    warnings: FrozenSet[str],
    color: str = "cyan",
) -> None:
    click.echo()
    default.display_section_name("STATISTICS")
    click.echo()
    error_count, warning_count = (
        counts["errors"],
        counts["warnings"],
    )
    click.secho(f"Total warnings: {warning_count}", fg=color)
    click.secho(f"Total errors: {error_count}", fg=color)
    for severity, check_stats in get_results_by_severity(event, warnings).items():
        if check_stats:
            click.echo()
            click.secho(severity, fg=color, bold=True, underline=True)
            # prints the check name and counts in descending order by count
            for check_name, count in sorted(
                check_stats.items(), key=lambda x: x[1], reverse=True
            ):
                click.secho(f"{check_name} : {count}", fg=color)
    click.echo()


def get_results_by_severity(
    event: events.Finished, warnings: FrozenSet[str]
) -> Dict[str, Dict[str, int]]:
    results: Dict[str, Dict[str, int]] = {"warnings": {}, "errors": {}}
    for check_name, result in event.total.items():
        if check_name in warnings and Status.failure in result:
            results["warnings"][check_name] = result[Status.failure]
        elif check_name not in warnings and Status.failure in result:
            results["errors"][check_name] = result[Status.failure]
    return results


def get_summary_counts(
    event: events.Finished, warnings: FrozenSet[str]
) -> Dict[str, int]:
    return {
        "successes": sum(
            results[Status.success]
            for check_name, results in event.total.items()
            if Status.success in results
        ),
        "warnings": sum(
            results[Status.failure]
            for check_name, results in event.total.items()
            if check_name in warnings and Status.failure in results
        ),
        "errors": sum(
            results[Status.failure]
            for check_name, results in event.total.items()
            if check_name not in warnings and Status.failure in results
        ),
        "exceptions": event.errored_count,
    }


def get_summary_output(
    counts: Dict[str, int], event: events.Finished, warnings: FrozenSet[str]
) -> Tuple[str, str, int]:
    counts_as_messages = [f"{num} {key}" for key, num in counts.items() if num > 0]
    if not counts_as_messages:
        message = "No checks performed."
        color = "yellow"
        status_code = 0
    else:
        message = f'{", ".join(counts_as_messages)} in {event.running_time:.2f}s'
        if counts["errors"] > 0 or counts["exceptions"] > 0:
            color = "red"
            status_code = 1
        else:
            color = "green"
            status_code = 0
    return message, color, status_code


def contains_warning(result: SerializedTestResult, warnings: FrozenSet[str]) -> bool:
    """Returns true if any check is result is a warning."""
    return any(_warn(check, warnings) for check in result.checks)


def get_unique_errors(
    checks: List[SerializedCheck], warnings: FrozenSet[str]
) -> List[SerializedCheck]:
    """Return only unique checks that should be displayed in the output."""
    seen: Set[Tuple[str, Optional[str]]] = set()
    unique_checks = []
    for check in reversed(checks):
        # There are also could be checks that didn't fail
        if (
            check.example is not None
            and check.value == Status.failure
            and not _warn(check, warnings)
            and (check.name, check.message) not in seen
        ):
            unique_checks.append(check)
            seen.add((check.name, check.message))
    return unique_checks


def get_unique_warnings(
    checks: List[SerializedCheck], warnings: FrozenSet[str]
) -> List[SerializedCheck]:
    """Return only unique checks that should be displayed in the output."""
    seen: Set[Tuple[str, Optional[str]]] = set()
    unique_checks = []
    for check in reversed(checks):
        # There are also could be checks that didn't fail
        if (
            check.example is not None
            and _warn(check, warnings)
            and (check.name, check.message) not in seen
        ):
            unique_checks.append(check)
            seen.add((check.name, check.message))
    return unique_checks


def make_verbose_name(attribute: str) -> str:
    return attribute.capitalize().replace("_", " ")


def warn_or_success(result: SerializedTestResult, warnings: FrozenSet[str]) -> bool:
    """All checks must be failures marked as warnings or successes.

    This function will return True if all checks are successes.
    """
    return all(
        check.value == Status.success or _warn(check, warnings) for check in result.checks
    )


def _warn(check: SerializedCheck, warnings: FrozenSet[str]) -> bool:
    return check.name in warnings and check.value == Status.failure


class OutputHandler(EventHandler):
    def __init__(self, warn: FrozenSet[str], statistics: bool) -> None:
        self.warn: FrozenSet[str] = warn
        self.statistics = statistics

    def handle_event(
        self, context: ExecutionContext, event: events.ExecutionEvent
    ) -> None:
        """Choose and execute a proper handler for the given event."""
        if isinstance(event, events.Initialized):
            default.handle_initialized(context, event)
        if isinstance(event, events.BeforeExecution):
            default.handle_before_execution(context, event)
        if isinstance(event, events.AfterExecution):
            context.hypothesis_output.extend(event.hypothesis_output)
            handle_after_execution(context, event, self.warn)
        if isinstance(event, events.Finished):
            handle_finished(context, event, self.warn, self.statistics)
        if isinstance(event, events.Interrupted):
            default.handle_interrupted(context, event)
        if isinstance(event, events.InternalError):
            handle_internal_error(context, event)



from typing import Callable, Dict, FrozenSet, Iterable, List, Optional, Tuple, Union
import os

import click
import hypothesis
from ibm_cloud_sdk_core.authenticators.iam_authenticator import IAMAuthenticator

from requests.models import Response
from schemathesis.cli.context import ExecutionContext
from schemathesis.cli.handlers import EventHandler
from schemathesis.cli.output.default import DefaultOutputStyleHandler
from schemathesis.cli.output.short import ShortOutputStyleHandler
from schemathesis.cli.options import CSVOption, OptionalInt, NotSet
from schemathesis.cli import callbacks, execute
from schemathesis.cli import replay as _replay
from schemathesis import loaders, runner

from schemathesis.hooks import GLOBAL_HOOK_DISPATCHER, HookContext
from schemathesis.types import Filter
from schemathesis import checks as checks_module
from schemathesis.models import Case

from ibm_service_validator.cli.handlers.output_handler import OutputHandler
from ibm_service_validator.handbook_rules import ADD_CASE_HOOKS, HANDBOOK_RULES
from ibm_service_validator.cli.process_config import (
    create_default_config,
    process_config,
)

ALL_CHECKS: tuple = checks_module.ALL_CHECKS + HANDBOOK_RULES
API_KEY = "IBM_CLOUD_SERVICE_VALIDATOR_API_KEY"
IAM_ENDPOINT = "IBM_CLOUD_SERVICE_VALIDATOR_IAM_ENDPOINT"
DEFAULT_WORKERS: int = 1


@click.group()
@click.option(
    "--set-api-key", "api_key", type=str, help="Set API key as an environment variable."
)
@click.option(
    "--set-iam-endpoint",
    "iam_endpoint",
    type=str,
    help="Set the IAM endpoint as an environment variable.",
)
def ibm_service_validator(api_key: str = str(), iam_endpoint: str = str()) -> None:
    set_environment_variable(api_key, API_KEY)
    set_environment_variable(iam_endpoint, IAM_ENDPOINT)


def set_environment_variable(val: str, env_var_name: str) -> None:
    if val:
        os.environ[env_var_name] = val


@ibm_service_validator.command(short_help="Run a suite of tests.")
@click.argument("schema", type=str, callback=callbacks.validate_schema)
@click.option(
    "--auth",
    "-a",
    type=str,
    callback=callbacks.validate_auth,
    help="Server user and password. Example: USER:PASSWORD",
)
@click.option(
    "--auth-type",
    "-A",
    type=click.Choice(["basic", "digest"], case_sensitive=False),
    default="basic",
    help="The authentication mechanism to be used. Defaults to 'basic'.",
)
@click.option(
    "--base-url",
    "-b",
    callback=callbacks.validate_base_url,
    help="The base-url of the API.",
)
@click.option(
    "--checks",
    "-c",
    type=str,
    default="",
    callback=lambda _, __, s: [c.strip() for c in s.split(",") if c.strip()],
    help="Comma-separated list of checks to run.",
)
@click.option(
    "--header",
    "-H",
    "headers",
    multiple=True,
    type=str,
    callback=callbacks.validate_headers,
    help="Custom header will be used in all requests to server. Ex: Authorization: Bearer 123",
)
@click.option(
    "--endpoint",
    "-E",
    "endpoints",
    type=str,
    multiple=True,
    callback=callbacks.validate_regex,
    help="Filter schemathesis test by endpoint pattern.",
)
@click.option(
    "--exitfirst",
    "-x",
    "exit_first",
    is_flag=True,
    default=False,
    help="Exit instantly on first error or failed test.",
)
@click.option(
    "--hypothesis-deadline",
    # max value to avoid overflow. It is maximum amount of days in milliseconds.
    type=OptionalInt(1, 999999999 * 24 * 3600 * 1000),
    help="Duration in milliseconds each individual example is not allowed to exceed.",
)
@click.option(
    "--hypothesis-derandomize",
    help="Use Hypothesis's deterministic mode.",
    is_flag=True,
    default=None,
)
@click.option(
    "--hypothesis-max-examples",
    type=click.IntRange(1),
    help="Maximum number of generated examples per each method/endpoint combination.",
)
@click.option(
    "--hypothesis-phases",
    default="explicit",
    type=CSVOption(hypothesis.Phase),
    help="Control which phases should be run.",
)
@click.option(
    "--hypothesis-seed", type=int, help="Set a seed to use for all Hypothesis tests."
)
@click.option(
    "--hypothesis-verbosity",
    type=click.Choice([item.name for item in hypothesis.Verbosity]),
    callback=callbacks.convert_verbosity,
    help="Verbosity level of Hypothesis messages.",
)
@click.option(
    "--method",
    "-M",
    "methods",
    type=str,
    multiple=True,
    callback=callbacks.validate_regex,
    help="Filter schemathesis test by HTTP method.",
)
@click.option(
    "--no-additional-cases",
    is_flag=True,
    default=False,
    help="Additional requests that target specific API behavior will not be sent.",
)
@click.option(
    "--request-timeout",
    type=click.IntRange(1),
    help="Timeout in milliseconds for network requests during the test run.",
)
@click.option(
    "--show-exception-tracebacks",
    is_flag=True,
    default=False,
    help="Show full tracebacks for internal exceptions.",
)
@click.option(
    "--statistics",
    "-s",
    is_flag=True,
    default=False,
    help="Show statistical summary of errors.",
)
@click.option(
    "--store-request-log",
    help="Store requests and responses into a file.",
    type=click.File("w"),
)
@click.option(
    "--tag",
    "-T",
    "tags",
    type=str,
    multiple=True,
    callback=callbacks.validate_regex,
    help="Filter schemathesis test by schema tag pattern.",
)
@click.option(
    "--validate-schema",
    help="Enable or disable validation of input schema.",
    type=bool,
    default=True,
)
@click.option("--verbosity", "-v", help="Control verbosity of output.", count=True)
@click.option(
    "--with-bearer", "-B", is_flag=True, help="Flag to send bearer token with requests."
)
def run(  # pylint: disable=too-many-arguments
    schema: str,
    auth: Optional[Tuple[str, str]],
    auth_type: str,
    base_url: Optional[str],
    checks: Optional[List[str]],
    headers: Dict[str, str],
    hypothesis_phases: Optional[List[hypothesis.Phase]],
    endpoints: Optional[Filter] = None,
    exit_first: bool = False,
    hypothesis_deadline: Optional[Union[int, NotSet]] = None,
    hypothesis_derandomize: Optional[bool] = None,
    hypothesis_max_examples: Optional[int] = None,
    hypothesis_seed: Optional[int] = None,
    hypothesis_verbosity: Optional[hypothesis.Verbosity] = None,
    methods: Optional[Filter] = None,
    no_additional_cases: bool = False,
    request_timeout: Optional[int] = None,
    show_exception_tracebacks: bool = False,
    statistics: bool = False,
    store_request_log: Optional[click.utils.LazyFile] = None,
    tags: Optional[Filter] = None,
    validate_schema: bool = True,
    verbosity: int = 0,
    with_bearer: bool = False,
) -> None:
    # pylint: disable=too-many-locals

    if with_bearer:
        if "Authorization" not in headers and "authorization" not in headers:
            headers["Authorization"] = get_bearer_token()
        else:
            raise click.UsageError(
                "--with-bearer flag used but Authorization header provided with --header."
            )

    on, warnings = (frozenset(checks), frozenset()) if checks else process_config()

    selected_checks = get_selected_checks(on)
    register_output_handler(warnings, statistics)
    if not no_additional_cases:
        register_add_case_hooks(on)

    # Invoke Schemathesis
    prepared_runner = runner.prepare(
        schema,
        app=None,
        auth=auth,
        auth_type=auth_type,
        base_url=base_url,
        checks=selected_checks,
        endpoint=endpoints,
        exit_first=exit_first,
        headers=headers,
        loader=loaders.from_path,
        method=methods,
        request_timeout=request_timeout,
        seed=hypothesis_seed,
        store_interactions=store_request_log is not None,
        tag=tags,
        validate_schema=validate_schema,
        workers_num=DEFAULT_WORKERS,
        hypothesis_deadline=hypothesis_deadline,
        hypothesis_derandomize=hypothesis_derandomize,
        hypothesis_max_examples=hypothesis_max_examples,
        hypothesis_phases=hypothesis_phases,
        hypothesis_report_multiple_bugs=False,
        hypothesis_suppress_health_check=None,
        hypothesis_verbosity=hypothesis_verbosity,
    )
    execute(
        prepared_runner,
        DEFAULT_WORKERS,
        show_exception_tracebacks,
        store_request_log,
        None,
        verbosity,
    )


def get_selected_checks(
    on: FrozenSet[str],
) -> Iterable[Callable[[Response, Case], None]]:
    return tuple(check for check in ALL_CHECKS if check.__name__ in on)


def register_output_handler(warnings: FrozenSet[str], statistics: bool) -> None:
    def after_init_cli_run_handlers(
        context: HookContext,
        handlers: List[EventHandler],
        execution_context: ExecutionContext,
    ) -> None:
        # filters out the Schemathesis output handlers and adds OutputHandler
        handlers[:] = [
            *filter(
                lambda handler: not isinstance(handler, DefaultOutputStyleHandler)
                and not isinstance(handler, ShortOutputStyleHandler),
                handlers,
            ),
            OutputHandler(warnings, statistics),
        ]

    GLOBAL_HOOK_DISPATCHER.register(after_init_cli_run_handlers)


def register_add_case_hooks(on: FrozenSet) -> None:
    # pylint: disable=bad-str-strip-call
    add_case_prefix = "add_"
    for case_hook in ADD_CASE_HOOKS:
        # add add_case hook if its corresponding check is on
        if case_hook.__name__.lstrip(add_case_prefix) in on:
            GLOBAL_HOOK_DISPATCHER.register_hook_with_name(case_hook, "add_case")


def get_bearer_token() -> str:
    if API_KEY not in os.environ or IAM_ENDPOINT not in os.environ:
        raise click.UsageError(
            f"Must set {API_KEY} and {IAM_ENDPOINT} environment variables to use --with-bearer."
        )

    try:
        bearer_token = IAMAuthenticator(
            os.environ[API_KEY], url=os.environ[IAM_ENDPOINT]
        ).token_manager.get_token()
        return "Bearer {0}".format(bearer_token)
    except Exception as e:  # pragma: no cover
        raise RuntimeError("Problem getting bearer token.") from e  # pragma: no cover


@ibm_service_validator.command(short_help="Create a default config file.")
@click.option(
    "--overwrite",
    "-o",
    "overwrite",
    is_flag=True,
    default=False,
    help="Overwrite an existing config file with the default config.",
)
@click.option(
    "--json",
    "-j",
    "write_json",
    is_flag=True,
    default=False,
    help="Overwrite an existing config file with the default config.",
)
def init(write_json: bool, overwrite: bool) -> None:
    create_default_config(write_json, overwrite)


@ibm_service_validator.command(short_help="Replay requests from a saved request log.")
@click.argument("cassette_path", type=click.Path(exists=True))
@click.option("--id", "id_", help="ID of request to replay.", type=str)
@click.option(
    "--status", help="Status (ERROR, FAILURE, SUCCESS) of requests to replay.", type=str
)
@click.option(
    "--uri", help="A regexp that filters requests by their request URI.", type=str
)
@click.option(
    "--method", help="A regexp that filters requests by their request method.", type=str,
)
@click.pass_context
def replay(  # pylint: disable=too-many-arguments
    context: click.Context,
    cassette_path: str,
    id_: Optional[str],
    status: Optional[str],
    uri: Optional[str],
    method: Optional[str],
) -> None:
    # pylint: disable=too-many-locals
    context.forward(_replay)



from setuptools import setup, find_packages

__version__ = "0.4.2"

with open("README.md", "r") as fh:
    readme = fh.read()

setup(
    name="ibm_service_validator",
    version=__version__,
    description="Python package, Schemathesis, extended to include IBM API Handbook validation.",
    url="https://github.com/IBM/service-validator",
    author_email="devexdev@us.ibm.com",
    license="Apache 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=["click", "ibm-cloud-sdk-core", "pyyaml", "schemathesis"],
    entry_points="""
        [console_scripts]
        ibm-service-validator=ibm_service_validator.cli:ibm_service_validator
    """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
